@@ -28,14 +28,13 @@ exception Already_bound
 
 type error =
     Unbound_type_variable of string
-  | Unbound_type_constructor of Longident.t
-  | Unbound_type_constructor_2 of Path.t
+  | Undefined_type_constructor of Path.t
   | Type_arity_mismatch of Longident.t * int * int
   | Bound_type_variable of string
   | Recursive_type
   | Unbound_row_variable of Longident.t
-  | Type_mismatch of (type_expr * type_expr) list
-  | Alias_type_mismatch of (type_expr * type_expr) list
+  | Type_mismatch of Ctype.Unification_trace.t
+  | Alias_type_mismatch of Ctype.Unification_trace.t
   | Present_has_conjunction of string
   | Present_has_no_type of string
   | Constructor_mismatch of type_expr * type_expr
@@ -45,167 +44,16 @@ type error =
   | Cannot_quantify of string * type_expr
   | Multiple_constraints_on_type of Longident.t
   | Method_mismatch of string * type_expr * type_expr
-  | Unbound_value of Longident.t
-  | Unbound_constructor of Longident.t
-  | Unbound_label of Longident.t
-  | Unbound_module of Longident.t
-  | Unbound_class of Longident.t
-  | Unbound_modtype of Longident.t
-  | Unbound_cltype of Longident.t
-  | Ill_typed_functor_application
-      of Longident.t * Longident.t * Includemod.error list option
-  | Illegal_reference_to_recursive_module
-  | Wrong_use_of_module of Longident.t * [ `Structure_used_as_functor
-                                         | `Abstract_used_as_functor
-                                         | `Functor_used_as_structure
-                                         | `Abstract_used_as_structure
-                                         | `Generative_used_as_applicative
-                                         ]
-  | Cannot_scrape_alias of Longident.t * Path.t
   | Opened_object of Path.t option
   | Not_an_object of type_expr
 
 exception Error of Location.t * Env.t * error
 exception Error_forward of Location.error
 
+(** Map indexed by type variable names. *)
+module TyVarMap = Misc.Stdlib.String.Map
 
-type variable_context = int * (string, type_expr) Tbl.t
-
-(* Narrowing unbound identifier errors. *)
-
-let rec narrow_unbound_lid_error : 'a. _ -> _ -> _ -> _ -> 'a =
-  fun env loc lid make_error ->
-  let check_module mlid =
-    try ignore (Env.lookup_module ~load:true mlid env) with
-    | Not_found ->
-        narrow_unbound_lid_error env loc mlid (fun lid -> Unbound_module lid)
-    | Env.Recmodule ->
-        raise (Error (loc, env, Illegal_reference_to_recursive_module))
-  in
-  let error e = raise (Error (loc, env, e)) in
-  begin match lid with
-  | Longident.Lident _ -> ()
-  | Longident.Ldot (mlid, _) ->
-      check_module mlid;
-      let md = Env.find_module (Env.lookup_module ~load:true mlid env) env in
-      begin match Env.scrape_alias env md.md_type with
-      | Mty_functor _ ->
-         error (Wrong_use_of_module (mlid, `Functor_used_as_structure))
-      | Mty_ident _ ->
-         error (Wrong_use_of_module (mlid, `Abstract_used_as_structure))
-      | Mty_alias(_, p) -> error (Cannot_scrape_alias(mlid, p))
-      | Mty_signature _ -> ()
-      end
-  | Longident.Lapply (flid, mlid) ->
-      check_module flid;
-      let fmd = Env.find_module (Env.lookup_module ~load:true flid env) env in
-      let mty_param =
-        match Env.scrape_alias env fmd.md_type with
-        | Mty_signature _ ->
-           error (Wrong_use_of_module (flid, `Structure_used_as_functor))
-        | Mty_ident _ ->
-           error (Wrong_use_of_module (flid, `Abstract_used_as_functor))
-        | Mty_alias(_, p) -> error (Cannot_scrape_alias(flid, p))
-        | Mty_functor (_, None, _) ->
-           error (Wrong_use_of_module (flid, `Generative_used_as_applicative))
-        | Mty_functor (_, Some mty_param, _) -> mty_param
-      in
-      check_module mlid;
-      let mpath = Env.lookup_module ~load:true mlid env in
-      let mmd = Env.find_module mpath env in
-      begin match Env.scrape_alias env mmd.md_type with
-      | Mty_alias(_, p) -> error (Cannot_scrape_alias(mlid, p))
-      | mty_arg ->
-         let details =
-           try Includemod.check_modtype_inclusion
-                 ~loc env mty_arg mpath mty_param;
-               None (* should be impossible *)
-           with Includemod.Error e -> Some e
-         in
-         error (Ill_typed_functor_application (flid, mlid, details))
-      end
-  end;
-  error (make_error lid)
-
-let find_component (lookup : ?loc:_ -> ?mark:_ -> _) make_error env loc lid =
-  try
-    match lid with
-    | Longident.Ldot (Longident.Lident "*predef*", s) ->
-        lookup ~loc (Longident.Lident s) Env.initial_safe_string
-    | _ ->
-        lookup ~loc lid env
-  with Not_found ->
-    narrow_unbound_lid_error env loc lid make_error
-  | Env.Recmodule ->
-    raise (Error (loc, env, Illegal_reference_to_recursive_module))
-
-let find_type env loc lid =
-  let path =
-    find_component Env.lookup_type (fun lid -> Unbound_type_constructor lid)
-      env loc lid
-  in
-  let decl = Env.find_type path env in
-  Builtin_attributes.check_deprecated loc decl.type_attributes (Path.name path);
-  (path, decl)
-
-let find_constructor =
-  find_component Env.lookup_constructor (fun lid -> Unbound_constructor lid)
-let find_all_constructors =
-  find_component Env.lookup_all_constructors
-    (fun lid -> Unbound_constructor lid)
-let find_label =
-  find_component Env.lookup_label (fun lid -> Unbound_label lid)
-let find_all_labels =
-  find_component Env.lookup_all_labels (fun lid -> Unbound_label lid)
-
-let find_class env loc lid =
-  let (path, decl) as r =
-    find_component Env.lookup_class (fun lid -> Unbound_class lid) env loc lid
-  in
-  Builtin_attributes.check_deprecated loc decl.cty_attributes (Path.name path);
-  r
-
-let find_value env loc lid =
-  Env.check_value_name (Longident.last lid) loc;
-  let (path, decl) as r =
-    find_component Env.lookup_value (fun lid -> Unbound_value lid) env loc lid
-  in
-  Builtin_attributes.check_deprecated loc decl.val_attributes (Path.name path);
-  r
-
-let lookup_module ?(load=false) env loc lid =
-  find_component (fun ?loc ?mark lid env -> (Env.lookup_module ~load ?loc ?mark lid env))
-    (fun lid -> Unbound_module lid) env loc lid
-
-let find_module env loc lid =
-  let path = lookup_module ~load:true env loc lid in
-  let decl = Env.find_module path env in
-  (* No need to check for deprecated here, this is done in Env. *)
-  (path, decl)
-
-let find_modtype env loc lid =
-  let (path, decl) as r =
-    find_component Env.lookup_modtype (fun lid -> Unbound_modtype lid)
-      env loc lid
-  in
-  Builtin_attributes.check_deprecated loc decl.mtd_attributes (Path.name path);
-  r
-
-let find_class_type env loc lid =
-  let (path, decl) as r =
-    find_component Env.lookup_cltype (fun lid -> Unbound_cltype lid)
-      env loc lid
-  in
-  Builtin_attributes.check_deprecated loc decl.clty_attributes (Path.name path);
-  r
-
-let unbound_constructor_error env lid =
-  narrow_unbound_lid_error env lid.loc lid.txt
-    (fun lid -> Unbound_constructor lid)
-
-let unbound_label_error env lid =
-  narrow_unbound_lid_error env lid.loc lid.txt
-    (fun lid -> Unbound_label lid)
+type variable_context = int * type_expr TyVarMap.t
 
 (* Support for first-class modules. *)
 
@@ -240,15 +88,15 @@ let create_package_mty fake loc env (p, l) =
 
 (* Translation of type expressions *)
 
-let type_variables = ref (Tbl.empty : (string, type_expr) Tbl.t)
+let type_variables = ref (TyVarMap.empty : type_expr TyVarMap.t)
 let univars        = ref ([] : (string * type_expr) list)
 let pre_univars    = ref ([] : type_expr list)
-let used_variables = ref (Tbl.empty : (string, type_expr * Location.t) Tbl.t)
+let used_variables = ref (TyVarMap.empty : (type_expr * Location.t) TyVarMap.t)
 
 let reset_type_variables () =
   reset_global_level ();
   Ctype.reset_reified_var_counter ();
-  type_variables := Tbl.empty
+  type_variables := TyVarMap.empty
 
 let narrow () =
   (increase_global_level (), !type_variables)
@@ -271,10 +119,13 @@ let newvar ?name () =
 
 let type_variable loc name =
   try
-    Tbl.find name !type_variables
+    TyVarMap.find name !type_variables
   with Not_found ->
     raise(Error(loc, Env.empty, Unbound_type_variable ("'" ^ name)))
 
+let valid_tyvar_name name =
+  name <> "" && name.[0] <> '_'
+
 let transl_type_param env styp =
   let loc = styp.ptyp_loc in
   match styp.ptyp_desc with
@@ -285,13 +136,13 @@ let transl_type_param env styp =
   | Ptyp_var name ->
       let ty =
         try
-          if name <> "" && name.[0] = '_' then
+          if not (valid_tyvar_name name) then
             raise (Error (loc, Env.empty, Invalid_variable_name ("'" ^ name)));
-          ignore (Tbl.find name !type_variables);
+          ignore (TyVarMap.find name !type_variables);
           raise Already_bound
         with Not_found ->
           let v = new_global_var ~name () in
-            type_variables := Tbl.add name v !type_variables;
+            type_variables := TyVarMap.add name v !type_variables;
             v
       in
         { ctyp_desc = Ttyp_var name; ctyp_type = ty; ctyp_env = env;
@@ -308,10 +159,6 @@ let transl_type_param env styp =
 let new_pre_univar ?name () =
   let v = newvar ?name () in pre_univars := v :: !pre_univars; v
 
-let rec swap_list = function
-    x :: y :: l -> y :: x :: swap_list l
-  | l -> l
-
 type policy = Fixed | Extensible | Univars
 
 let rec transl_type env policy styp =
@@ -335,17 +182,17 @@ and transl_type_aux env policy styp =
       ctyp Ttyp_any ty
   | Ptyp_var name ->
     let ty =
-      if name <> "" && name.[0] = '_' then
+      if not (valid_tyvar_name name) then
         raise (Error (styp.ptyp_loc, env, Invalid_variable_name ("'" ^ name)));
       begin try
         instance (List.assoc name !univars)
       with Not_found -> try
-        instance (fst(Tbl.find name !used_variables))
+        instance (fst (TyVarMap.find name !used_variables))
       with Not_found ->
         let v =
           if policy = Univars then new_pre_univar ~name () else newvar ~name ()
         in
-        used_variables := Tbl.add name (v, styp.ptyp_loc) !used_variables;
+        used_variables := TyVarMap.add name (v, styp.ptyp_loc) !used_variables;
         v
       end
     in
@@ -366,7 +213,7 @@ and transl_type_aux env policy styp =
     let ty = newty (Ttuple (List.map (fun ctyp -> ctyp.ctyp_type) ctys)) in
     ctyp (Ttyp_tuple ctys) ty
   | Ptyp_constr(lid, stl) ->
-      let (path, decl) = find_type env lid.loc lid.txt in
+      let (path, decl) = Env.lookup_type ~loc:lid.loc lid.txt env in
       let stl =
         match stl with
         | [ {ptyp_desc=Ptyp_any} as t ] when decl.type_arity > 1 ->
@@ -388,7 +235,9 @@ and transl_type_aux env policy styp =
       List.iter2
         (fun (sty, cty) ty' ->
            try unify_param env ty' cty.ctyp_type with Unify trace ->
-             raise (Error(sty.ptyp_loc, env, Type_mismatch (swap_list trace))))
+             let trace = Unification_trace.swap trace in
+             raise (Error(sty.ptyp_loc, env, Type_mismatch trace))
+        )
         (List.combine stl args) params;
       let constr =
         newconstr path (List.map (fun ctyp -> ctyp.ctyp_type) args) in
@@ -404,8 +253,7 @@ and transl_type_aux env policy styp =
   | Ptyp_class(lid, stl) ->
       let (path, decl, _is_variant) =
         try
-          let path = Env.lookup_type lid.txt env in
-          let decl = Env.find_type path env in
+          let path, decl = Env.find_type_by_name lid.txt env in
           let rec check decl =
             match decl.type_manifest with
               None -> raise Not_found
@@ -418,6 +266,7 @@ and transl_type_aux env policy styp =
           in check decl;
           Location.deprecated styp.ptyp_loc
             "old syntax for polymorphic variant type";
+          ignore(Env.lookup_type ~loc:lid.loc lid.txt env);
           (path, decl,true)
         with Not_found -> try
           let lid2 =
@@ -426,11 +275,11 @@ and transl_type_aux env policy styp =
             | Longident.Ldot(r, s)   -> Longident.Ldot (r, "#" ^ s)
             | Longident.Lapply(_, _) -> fatal_error "Typetexp.transl_type"
           in
-          let path = Env.lookup_type lid2 env in
-          let decl = Env.find_type path env in
+          let path, decl = Env.find_type_by_name lid2 env in
+          ignore(Env.lookup_cltype ~loc:lid.loc lid.txt env);
           (path, decl, false)
         with Not_found ->
-          ignore (find_class env lid.loc lid.txt); assert false
+          ignore (Env.lookup_cltype ~loc:lid.loc lid.txt env); assert false
       in
       if List.length stl <> decl.type_arity then
         raise(Error(styp.ptyp_loc, env,
@@ -441,7 +290,9 @@ and transl_type_aux env policy styp =
       List.iter2
         (fun (sty, cty) ty' ->
            try unify_var env ty' cty.ctyp_type with Unify trace ->
-             raise (Error(sty.ptyp_loc, env, Type_mismatch (swap_list trace))))
+             let trace = Unification_trace.swap trace in
+             raise (Error(sty.ptyp_loc, env, Type_mismatch trace))
+        )
         (List.combine stl args) params;
         let ty_args = List.map (fun ctyp -> ctyp.ctyp_type) args in
       let ty =
@@ -465,7 +316,7 @@ and transl_type_aux env policy styp =
           in
           let row = { row_closed = true; row_fields = fields;
                       row_bound = (); row_name = Some (path, ty_args);
-                      row_fixed = false; row_more = newvar () } in
+                      row_fixed = None; row_more = newvar () } in
           let static = Btype.static_row row in
           let row =
             if static then { row with row_more = newty Tnil }
@@ -487,21 +338,22 @@ and transl_type_aux env policy styp =
           let t =
             try List.assoc alias !univars
             with Not_found ->
-              instance (fst(Tbl.find alias !used_variables))
+              instance (fst(TyVarMap.find alias !used_variables))
           in
           let ty = transl_type env policy st in
           begin try unify_var env t ty.ctyp_type with Unify trace ->
-            let trace = swap_list trace in
+            let trace = Unification_trace.swap trace in
             raise(Error(styp.ptyp_loc, env, Alias_type_mismatch trace))
           end;
           ty
         with Not_found ->
           if !Clflags.principal then begin_def ();
           let t = newvar () in
-          used_variables := Tbl.add alias (t, styp.ptyp_loc) !used_variables;
+          used_variables :=
+            TyVarMap.add alias (t, styp.ptyp_loc) !used_variables;
           let ty = transl_type env policy st in
           begin try unify_var env t ty.ctyp_type with Unify trace ->
-            let trace = swap_list trace in
+            let trace = Unification_trace.swap trace in
             raise(Error(styp.ptyp_loc, env, Alias_type_mismatch trace))
           end;
           if !Clflags.principal then begin
@@ -511,8 +363,8 @@ and transl_type_aux env policy styp =
           let t = instance t in
           let px = Btype.proxy t in
           begin match px.desc with
-          | Tvar None -> Btype.log_type px; px.desc <- Tvar (Some alias)
-          | Tunivar None -> Btype.log_type px; px.desc <- Tunivar (Some alias)
+          | Tvar None -> Btype.set_type_desc px (Tvar (Some alias))
+          | Tunivar None -> Btype.set_type_desc px (Tunivar (Some alias))
           | _ -> ()
           end;
           { ty with ctyp_type = t }
@@ -523,7 +375,7 @@ and transl_type_aux env policy styp =
       let mkfield l f =
         newty (Tvariant {row_fields=[l,f]; row_more=newvar();
                          row_bound=(); row_closed=true;
-                         row_fixed=false; row_name=None}) in
+                         row_fixed=None; row_name=None}) in
       let hfields = Hashtbl.create 17 in
       let add_typed_field loc l f =
         let h = Btype.hash_variant l in
@@ -539,11 +391,14 @@ and transl_type_aux env policy styp =
         with Not_found ->
           Hashtbl.add hfields h (l,f)
       in
-      let add_field = function
-          Rtag (l, attrs, c, stl) ->
+      let add_field field =
+        let rf_loc = field.prf_loc in
+        let rf_attributes = field.prf_attributes in
+        let rf_desc = match field.prf_desc with
+        | Rtag (l, c, stl) ->
             name := None;
             let tl =
-              Builtin_attributes.warning_scope attrs
+              Builtin_attributes.warning_scope rf_attributes
                 (fun () -> List.map (transl_type env policy) stl)
             in
             let f = match present with
@@ -559,7 +414,7 @@ and transl_type_aux env policy styp =
                       Rpresent (Some st.ctyp_type)
             in
             add_typed_field styp.ptyp_loc l.txt f;
-              Ttag (l,attrs,c,tl)
+              Ttag (l,c,tl)
         | Rinherit sty ->
             let cty = transl_type env policy sty in
             let ty = cty.ctyp_type in
@@ -568,20 +423,13 @@ and transl_type_aux env policy styp =
                 {desc=Tconstr(p, tl, _)} -> Some(p, tl)
               | _                        -> None
             in
-            begin try
-              (* Set name if there are no fields yet *)
-              Hashtbl.iter (fun _ _ -> raise Exit) hfields;
-              name := nm
-            with Exit ->
-              (* Unset it otherwise *)
-              name := None
-            end;
+            name := if Hashtbl.length hfields <> 0 then None else nm;
             let fl = match expand_head env cty.ctyp_type, nm with
               {desc=Tvariant row}, _ when Btype.static_row row ->
                 let row = Btype.row_repr row in
                 row.row_fields
             | {desc=Tvar _}, Some(p, _) ->
-                raise(Error(sty.ptyp_loc, env, Unbound_type_constructor_2 p))
+                raise(Error(sty.ptyp_loc, env, Undefined_type_constructor p))
             | _ ->
                 raise(Error(sty.ptyp_loc, env, Not_a_variant ty))
             in
@@ -602,6 +450,8 @@ and transl_type_aux env policy styp =
                 add_typed_field sty.ptyp_loc l f)
               fl;
               Tinherit cty
+        in
+        { rf_desc; rf_loc; rf_attributes; }
       in
       let tfields = List.map add_field fields in
       let fields = Hashtbl.fold (fun _ p l -> p :: l) hfields [] in
@@ -615,7 +465,7 @@ and transl_type_aux env policy styp =
       let row =
         { row_fields = List.rev fields; row_more = newvar ();
           row_bound = (); row_closed = (closed = Closed);
-          row_fixed = false; row_name = !name } in
+          row_fixed = None; row_name = !name } in
       let static = Btype.static_row row in
       let row =
         if static then { row with row_more = newty Tnil }
@@ -688,13 +538,16 @@ and transl_fields env policy o fields =
           raise(Error(loc, env, Method_mismatch (l, ty, ty')))
     with Not_found ->
       Hashtbl.add hfields l ty in
-  let add_field = function
-    | Otag (s, a, ty1) -> begin
+  let add_field {pof_desc; pof_loc; pof_attributes;} =
+    let of_loc = pof_loc in
+    let of_attributes = pof_attributes in
+    let of_desc = match pof_desc with
+    | Otag (s, ty1) -> begin
         let ty1 =
-          Builtin_attributes.warning_scope a
+          Builtin_attributes.warning_scope of_attributes
             (fun () -> transl_poly_type env policy ty1)
         in
-        let field = OTtag (s, a, ty1) in
+        let field = OTtag (s, ty1) in
         add_typed_field ty1.ctyp_loc s.txt ty1.ctyp_type;
         field
       end
@@ -720,9 +573,11 @@ and transl_fields env policy o fields =
             OTinherit cty
             end
         | {desc=Tvar _}, Some p ->
-            raise (Error (sty.ptyp_loc, env, Unbound_type_constructor_2 p))
+            raise (Error (sty.ptyp_loc, env, Undefined_type_constructor p))
         | _ -> raise (Error (sty.ptyp_loc, env, Not_an_object t))
       end in
+    { of_desc; of_loc; of_attributes; }
+  in
   let object_fields = List.map add_field fields in
   let fields = Hashtbl.fold (fun s ty l -> (s, ty) :: l) hfields [] in
   let ty_init =
@@ -743,9 +598,10 @@ let rec make_fixed_univars ty =
     match ty.desc with
     | Tvariant row ->
         let row = Btype.row_repr row in
-        if Btype.is_Tunivar (Btype.row_more row) then
+        let more = Btype.row_more row in
+        if Btype.is_Tunivar more then
           ty.desc <- Tvariant
-              {row with row_fixed=true;
+              {row with row_fixed=Some(Univar more);
                row_fields = List.map
                  (fun (s,f as p) -> match Btype.row_field_repr f with
                    Reither (c, tl, _m, r) -> s, Reither (c, tl, true, r)
@@ -764,21 +620,21 @@ let create_package_mty = create_package_mty false
 
 let globalize_used_variables env fixed =
   let r = ref [] in
-  Tbl.iter
+  TyVarMap.iter
     (fun name (ty, loc) ->
       let v = new_global_var () in
       let snap = Btype.snapshot () in
       if try unify env v ty; true with _ -> Btype.backtrack snap; false
       then try
-        r := (loc, v,  Tbl.find name !type_variables) :: !r
+        r := (loc, v,  TyVarMap.find name !type_variables) :: !r
       with Not_found ->
         if fixed && Btype.is_Tvar (repr ty) then
           raise(Error(loc, env, Unbound_type_variable ("'"^name)));
         let v2 = new_global_var () in
         r := (loc, v, v2) :: !r;
-        type_variables := Tbl.add name v2 !type_variables)
+        type_variables := TyVarMap.add name v2 !type_variables)
     !used_variables;
-  used_variables := Tbl.empty;
+  used_variables := TyVarMap.empty;
   fun () ->
     List.iter
       (function (loc, t1, t2) ->
@@ -787,23 +643,23 @@ let globalize_used_variables env fixed =
       !r
 
 let transl_simple_type env fixed styp =
-  univars := []; used_variables := Tbl.empty;
+  univars := []; used_variables := TyVarMap.empty;
   let typ = transl_type env (if fixed then Fixed else Extensible) styp in
   globalize_used_variables env fixed ();
   make_fixed_univars typ.ctyp_type;
   typ
 
 let transl_simple_type_univars env styp =
-  univars := []; used_variables := Tbl.empty; pre_univars := [];
+  univars := []; used_variables := TyVarMap.empty; pre_univars := [];
   begin_def ();
   let typ = transl_type env Univars styp in
   (* Only keep already global variables in used_variables *)
   let new_variables = !used_variables in
-  used_variables := Tbl.empty;
-  Tbl.iter
+  used_variables := TyVarMap.empty;
+  TyVarMap.iter
     (fun name p ->
-      if Tbl.mem name !type_variables then
-        used_variables := Tbl.add name p !used_variables)
+      if TyVarMap.mem name !type_variables then
+        used_variables := TyVarMap.add name p !used_variables)
     new_variables;
   globalize_used_variables env false ();
   end_def ();
@@ -823,10 +679,18 @@ let transl_simple_type_univars env styp =
         instance (Btype.newgenty (Tpoly (typ.ctyp_type, univs))) }
 
 let transl_simple_type_delayed env styp =
-  univars := []; used_variables := Tbl.empty;
+  univars := []; used_variables := TyVarMap.empty;
+  begin_def ();
   let typ = transl_type env Extensible styp in
+  end_def ();
   make_fixed_univars typ.ctyp_type;
-  (typ, globalize_used_variables env false)
+  (* This brings the used variables to the global level, but doesn't link them
+     to their other occurrences just yet. This will be done when [force] is
+     called. *)
+  let force = globalize_used_variables env false in
+  (* Generalizes everything except the variables that were just globalized. *)
+  generalize typ.ctyp_type;
+  (typ, instance typ.ctyp_type, force)
 
 let transl_type_scheme env styp =
   reset_type_variables();
@@ -842,39 +706,14 @@ let transl_type_scheme env styp =
 open Format
 open Printtyp
 
-let spellcheck ppf fold env lid =
-  let choices ~path name =
-    let env = fold (fun x xs -> x::xs) path env [] in
-    Misc.spellcheck env name in
-  match lid with
-    | Longident.Lapply _ -> ()
-    | Longident.Lident s ->
-       Misc.did_you_mean ppf (fun () -> choices ~path:None s)
-    | Longident.Ldot (r, s) ->
-       Misc.did_you_mean ppf (fun () -> choices ~path:(Some r) s)
-
-let fold_descr fold get_name f = fold (fun descr acc -> f (get_name descr) acc)
-let fold_simple fold4 f = fold4 (fun name _path _descr acc -> f name acc)
-
-let fold_values = fold_simple Env.fold_values
-let fold_types = fold_simple Env.fold_types
-let fold_modules = fold_simple Env.fold_modules
-let fold_constructors = fold_descr Env.fold_constructors (fun d -> d.cstr_name)
-let fold_labels = fold_descr Env.fold_labels (fun d -> d.lbl_name)
-let fold_classs = fold_simple Env.fold_classs
-let fold_modtypes = fold_simple Env.fold_modtypes
-let fold_cltypes = fold_simple Env.fold_cltypes
-
 let report_error env ppf = function
   | Unbound_type_variable name ->
-     (* we don't use "spellcheck" here: the function that raises this
-        error seems not to be called anywhere, so it's unclear how it
-        should be handled *)
-    fprintf ppf "Unbound type parameter %s@." name
-  | Unbound_type_constructor lid ->
-    fprintf ppf "Unbound type constructor %a" longident lid;
-    spellcheck ppf fold_types env lid;
-  | Unbound_type_constructor_2 p ->
+      let add_name name _ l = if name = "_" then l else ("'" ^ name) :: l in
+      let names = TyVarMap.fold add_name !type_variables [] in
+    fprintf ppf "The type variable %s is unbound in this type declaration.@ %a"
+      name
+      did_you_mean (fun () -> Misc.spellcheck names name )
+  | Undefined_type_constructor p ->
     fprintf ppf "The type constructor@ %a@ is not yet completely defined"
       path p
   | Type_arity_mismatch(lid, expected, provided) ->
@@ -883,7 +722,7 @@ let report_error env ppf = function
         but is here applied to %i argument(s)@]"
       longident lid expected provided
   | Bound_type_variable name ->
-    fprintf ppf "Already bound type parameter '%s" name
+    fprintf ppf "Already bound type parameter %a" Pprintast.tyvar name
   | Recursive_type ->
     fprintf ppf "This type is recursive"
   | Unbound_row_variable lid ->
@@ -905,17 +744,22 @@ let report_error env ppf = function
   | Present_has_conjunction l ->
       fprintf ppf "The present constructor %s has a conjunctive type" l
   | Present_has_no_type l ->
-      fprintf ppf "The present constructor %s has no type" l
+      fprintf ppf
+        "@[<v>@[The constructor %s is missing from the upper bound@ \
+         (between '<'@ and '>')@ of this polymorphic variant@ \
+         but is present in@ its lower bound (after '>').@]@,\
+         @[Hint: Either add `%s in the upper bound,@ \
+         or remove it@ from the lower bound.@]@]"
+         l l
   | Constructor_mismatch (ty, ty') ->
       wrap_printing_env ~error:true env (fun ()  ->
         Printtyp.reset_and_mark_loops_list [ty; ty'];
         fprintf ppf "@[<hov>%s %a@ %s@ %a@]"
           "This variant type contains a constructor"
-          Printtyp.type_expr ty
+          !Oprint.out_type (tree_of_typexp false ty)
           "which should be"
-          Printtyp.type_expr ty')
+           !Oprint.out_type (tree_of_typexp false ty'))
   | Not_a_variant ty ->
-      Printtyp.reset_and_mark_loops ty;
       fprintf ppf
         "@[The type %a@ does not expand to a polymorphic variant type@]"
         Printtyp.type_expr ty;
@@ -933,70 +777,21 @@ let report_error env ppf = function
       fprintf ppf "The type variable name %s is not allowed in programs" name
   | Cannot_quantify (name, v) ->
       fprintf ppf
-        "@[<hov>The universal type variable '%s cannot be generalized:@ %s.@]"
-        name
-        (if Btype.is_Tvar v then "it escapes its scope" else
-         if Btype.is_Tunivar v then "it is already bound to another variable"
-         else "it is not a variable")
+        "@[<hov>The universal type variable %a cannot be generalized:@ "
+        Pprintast.tyvar name;
+      if Btype.is_Tvar v then
+        fprintf ppf "it escapes its scope"
+      else if Btype.is_Tunivar v then
+        fprintf ppf "it is already bound to another variable"
+      else
+        fprintf ppf "it is bound to@ %a" Printtyp.type_expr v;
+      fprintf ppf ".@]";
   | Multiple_constraints_on_type s ->
       fprintf ppf "Multiple constraints for type %a" longident s
   | Method_mismatch (l, ty, ty') ->
       wrap_printing_env ~error:true env (fun ()  ->
-        Printtyp.reset_and_mark_loops_list [ty; ty'];
         fprintf ppf "@[<hov>Method '%s' has type %a,@ which should be %a@]"
           l Printtyp.type_expr ty Printtyp.type_expr ty')
-  | Unbound_value lid ->
-      fprintf ppf "Unbound value %a" longident lid;
-      spellcheck ppf fold_values env lid;
-  | Unbound_module lid ->
-      fprintf ppf "Unbound module %a" longident lid;
-      spellcheck ppf fold_modules env lid;
-  | Unbound_constructor lid ->
-      fprintf ppf "Unbound constructor %a" longident lid;
-      spellcheck ppf fold_constructors env lid;
-  | Unbound_label lid ->
-      fprintf ppf "Unbound record field %a" longident lid;
-      spellcheck ppf fold_labels env lid;
-  | Unbound_class lid ->
-      fprintf ppf "Unbound class %a" longident lid;
-      spellcheck ppf fold_classs env lid;
-  | Unbound_modtype lid ->
-      fprintf ppf "Unbound module type %a" longident lid;
-      spellcheck ppf fold_modtypes env lid;
-  | Unbound_cltype lid ->
-      fprintf ppf "Unbound class type %a" longident lid;
-      spellcheck ppf fold_cltypes env lid;
-  | Ill_typed_functor_application (flid, mlid, details) ->
-     (match details with
-     | None ->
-        fprintf ppf "@[Ill-typed functor application %a(%a)@]"
-          longident flid longident mlid
-     | Some inclusion_error ->
-        fprintf ppf "@[The type of %a does not match %a's parameter@\n%a@]"
-          longident mlid longident flid Includemod.report_error inclusion_error)
-  | Illegal_reference_to_recursive_module ->
-     fprintf ppf "Illegal recursive module reference"
-  | Wrong_use_of_module (lid, details) ->
-     (match details with
-     | `Structure_used_as_functor ->
-        fprintf ppf "@[The module %a is a structure, it cannot be applied@]"
-          longident lid
-     | `Abstract_used_as_functor ->
-        fprintf ppf "@[The module %a is abstract, it cannot be applied@]"
-          longident lid
-     | `Functor_used_as_structure ->
-        fprintf ppf "@[The module %a is a functor, \
-                       it cannot have any components@]" longident lid
-     | `Abstract_used_as_structure ->
-        fprintf ppf "@[The module %a is abstract, \
-                       it cannot have any components@]" longident lid
-     | `Generative_used_as_applicative ->
-        fprintf ppf "@[The functor %a is generative,@ it@ cannot@ be@ \
-                       applied@ in@ type@ expressions@]" longident lid)
-  | Cannot_scrape_alias(lid, p) ->
-      fprintf ppf
-        "The module %a is an alias for module %a, which is missing"
-        longident lid path p
   | Opened_object nm ->
       fprintf ppf
         "Illegal open object type%a"
@@ -1004,7 +799,6 @@ let report_error env ppf = function
              Some p -> fprintf ppf "@ %a" path p
            | None -> fprintf ppf "") nm
   | Not_an_object ty ->
-      Printtyp.reset_and_mark_loops ty;
       fprintf ppf "@[The type %a@ is not an object type@]"
         Printtyp.type_expr ty
 
@@ -1012,7 +806,7 @@ let () =
   Location.register_error_of_exn
     (function
       | Error (loc, env, err) ->
-        Some (Location.error_of_printer loc (report_error env) err)
+        Some (Location.error_of_printer ~loc (report_error env) err)
       | Error_forward err ->
         Some err
       | _ ->
